"""
Programm 9.1 - Lösung der Poisson-Gleichung (pdenr_gb.py) auf einem Nichtrechteckgebiet mit Dirichlet-
               und Neumann-RB
"""

import numpy as np


def pdenr_gb():
    """
    Programm 9.1 - Lösung der Poisson-Gleichung (pdenr_gb.py) auf einem Nichtrechteckgebiet mit Dirichlet-
                   und Neumann-RB

    Aufruf:

        u, x, y, it = pdenr_gb()

    :return: Lösung u(x, y) auf dem Gitter x, y
    """
    # Materialdaten, Quellen
    la = 1.0
    q = 1.0
    ff = 0.0

    # Gebeitslänge, Gebietshöhe
    ll = 1.0
    hh = 10.5*ll/16.0
    ll1 = ll/2.0
    hh1 = hh*3.0/7.0
    mm = (hh - hh1)/(ll1 - ll)
    nn = hh - mm*ll1

    # Dimensionierung des Gitters
    nf = 6  # nf = 3
    n = nf*8 + 1
    m = nf*7 + 1

    # Gitterinitialisierung
    x, y = np.zeros(shape=(n,)), np.zeros(shape=(m,))
    x[0], y[0] = 0.0, 0.0
    h = 1.0/(nf*8.0)
    k = 3.0*h/4.0
    for i in range(1, n):
        x[i] = i*h
    for j in range(1, m):
        y[j] = j*k

    # Initialisierung der Gebietsmatrix
    mg = np.ones(shape=(n, m), dtype=int)
    mg[:, -2] = 3  # set boundary
    mg[0, :] = 5  # set boundary
    mg[:, 0] = 7  # set boundary
    mg[-2, :] = 9  # set boundary
    mg[0, -2] = 4  # set corner
    mg[0, 0] = 6  # set corner
    mg[-2, 0] = 8  # set corner
    for i in range(n):
        for j in range(m):
            if abs(mm*x[i] + nn - y[j]) < 1.0e-06:
                mg[i, j] = 2
            if y[j] > mm*x[i] + nn:
                mg[i, j] = 0

    # Startfeld incl. homogener Dirichlet-Randbedingungen
    u = np.zeros(shape=(n, m))
    ualt = np.array(u, copy=True)

    # Konstanten
    omega = 1.65
    h2 = h**2
    k2 = k**2
    dd1 = 1.0/(2.0/h2 + 2.0/k2)
    dd2 = 1.0/(1.0/h2 + 1.0/k2)
    dd3 = 1.0/(2.0/h2 + 3.0/k2)
    dd4 = 1.0/(3.0/h2 + 3.0/k2)
    dd5 = 1.0/(3.0/h2 + 2.0/k2)
    dd6 = dd4
    dd7 = dd3
    dd8 = dd4
    dd9 = dd5
    whk = q*(((h2 + k2)**0.5)/(h*k))/la
    udiff = 1.0
    unorm = 1.0
    it = -1
    la2 = 2.0*la

    # Randwerte
    un = 0
    us = 0
    uo = 0
    uw = 0

    # sor-Iteration
    max_it: int = 1200
    for it in range(max_it):
        if (udiff/unorm)**0.5 <= 1.0e-05: break  # verlasse for-Schleife vorzeitig
        print(f'it: {it} bzw. {100.0*it/1200:.2f} % von maximal erlaubten Iterationen ({max_it}).')
        for i in range(n - 1):
            for j in range(m - 1):
                ualt[i, j] = u[i, j]
        for i in range(n - 1):
            for j in range(m - 1):
                _mg = mg[i, j]
                if _mg == 1:
                    u[i, j] = dd1*(ff/la + (u[i + 1, j] + u[i - 1, j])/h2 + (u[i, j + 1] + u[i, j - 1])/k2)
                elif _mg == 2:
                    u[i, j] = dd2*(ff/la2 + u[i - 1, j]/h2 + u[i, j - 1]/k2 + whk)
                elif _mg == 3:
                    u[i, j] = dd3*(ff/la + (u[i + 1, j] + u[i - 1, j])/h2 + (2.0*un + u[i, j - 1])/k2)
                elif _mg == 4:
                    u[i, j] = dd4*(ff/la + (u[i + 1, j] + 2.0*uw)/h2 + (2.0*un + u[i, j - 1])/k2)
                elif _mg == 5:
                    u[i, j] = dd5*(ff/la + (u[i + 1, j] + 2.0*uw)/h2 + (u[i, j + 1] + u[i, j - 1])/k2)
                elif _mg == 6:
                    u[i, j] = dd6*(ff/la + (u[i + 1, j] + 2.0*uw)/h2 + (u[i, j + 1] + 2.0*us)/k2)
                elif _mg == 7:
                    u[i, j] = dd7*(ff/la + (u[i + 1, j] + u[i - 1, j])/h2 + (u[i, j + 1] + 2.0*us)/k2)
                elif _mg == 8:
                    u[i, j] = dd8*(ff/la + (2.0*uo + u[i - 1, j])/h2 + (u[i, j + 1] + 2.0*us)/k2)
                elif _mg == 9:
                    u[i, j] = dd9*(ff/la + (2.0*uo + u[i - 1, j])/h2 + (u[i, j + 1] + u[i, j - 1])/k2)
        udiff = 0.0
        unorm = 0.0
        for i in range(n - 1):
            for j in range(m - 1):
                u[i, j] = omega*u[i, j] + (1.0 - omega)*ualt[i, j]
                udiff = udiff + (u[i, j] - ualt[i, j])**2
                unorm = unorm + ualt[i, j]**2
        if unorm == 0.0:
            unorm = 1.0
    u[mg == 0.0] = 0.0
    return u, x, y, it


def main():
    u, x, y, it = pdenr_gb()

    import numpy as np
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D

    X, Y = np.meshgrid(x, y, indexing='ij')
    fig = plt.figure()
    axes = plt.axes(projection='3d')

    Z = u
    plt.xlabel('x')
    plt.ylabel('y')

    axes.set_title('Surface plot of pde solution')
    axes.plot_surface(X, Y, Z)

    plt.show()


if __name__ == '__main__':
    main()
